#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from typing import Any

from keystoneauth1 import loading
from oslo_config import cfg

from oslo_limit._i18n import _

__all__ = [
    'list_opts',
    'register_opts',
]

CONF = cfg.CONF

_options = [
    cfg.StrOpt(
        'endpoint_id',
        help=_("The service's endpoint id which is registered in Keystone."),
    ),
    cfg.StrOpt(
        'endpoint_service_name', help=_("Service name for endpoint discovery")
    ),
    cfg.StrOpt(
        'endpoint_service_type', help=_("Service type for endpoint discovery")
    ),
    cfg.StrOpt(
        'endpoint_region_name', help=_("Region to which the endpoint belongs")
    ),
    cfg.StrOpt(
        'endpoint_interface',
        default='public',
        choices=[
            'public',
            'publicURL',
            'internal',
            'internalURL',
            'admin',
            'adminURL',
        ],
        help=_("The interface for endpoint discovery"),
    ),
]

_option_group = 'oslo_limit'


def list_opts() -> list[tuple[str, list[cfg.Opt]]]:
    """Return a list of oslo.config options available in the library.

    :returns: a list of (group_name, opts) tuples
    """

    return [
        (
            _option_group,
            copy.deepcopy(_options)
            + loading.get_session_conf_options()
            + loading.get_auth_plugin_conf_options('password')
            + loading.get_auth_plugin_conf_options('v2password')
            + loading.get_auth_plugin_conf_options('v3password')
            + loading.get_adapter_conf_options(include_deprecated=False),
        )
    ]


def register_opts(conf: cfg.ConfigOpts) -> None:
    loading.register_session_conf_options(CONF, _option_group)
    loading.register_adapter_conf_options(
        CONF, _option_group, include_deprecated=False
    )

    loading.register_auth_conf_options(CONF, _option_group)
    plugin_name = CONF.oslo_limit.auth_type
    if plugin_name:
        plugin_loader: loading.BaseLoader[Any]
        plugin_loader = loading.get_plugin_loader(plugin_name)
        plugin_opts = loading.get_auth_plugin_conf_options(plugin_loader)
        CONF.register_opts(plugin_opts, group=_option_group)
    conf.register_opts(_options, group=_option_group)



#    Copyright 2021 Red Hat, Inc
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Generator
from typing import Any
from unittest import mock

import fixtures as fixtures

from openstack.identity.v3 import endpoint as _endpoint
from openstack.identity.v3 import limit as _limit
from openstack.identity.v3 import region as _region
from openstack.identity.v3 import registered_limit as _registered_limit
from openstack.identity.v3 import service as _service
from openstack.test import fakes as sdk_fakes


def _fake_services(
    **query: dict[str, Any],
) -> Generator[_service.Service, None, None]:
    # we are the only ones calling this, so we know exactly what we should be
    # calling it with
    assert set(query) == {'type', 'name'}
    yield sdk_fakes.generate_fake_resource(
        _service.Service, type=query['type'], name=query['name']
    )


def _fake_region(
    region: str | _region.Region,
) -> _region.Region:
    if isinstance(region, str):
        region_id = region
    else:
        region_id = region.id

    return sdk_fakes.generate_fake_resource(_region.Region, id=region_id)


def _fake_endpoints(
    **query: dict[str, Any],
) -> Generator[_endpoint.Endpoint, None, None]:
    # we are the only ones calling this, so we know exactly what we should be
    # calling it with
    assert set(query) == {'service_id', 'region_id', 'interface'}
    yield sdk_fakes.generate_fake_resource(
        _endpoint.Endpoint,
        service_id=query['service_id'],
        region_id=query['region_id'],
        interface=query['interface'],
    )


class LimitFixture(fixtures.Fixture):
    def __init__(
        self,
        reglimits: dict[str, int],
        projlimits: dict[str, dict[str, int]],
    ) -> None:
        """A fixture for testing code that relies on Keystone Unified Limits.

        :param reglimits: A dictionary of {resource_name: limit} values to
                          simulate registered limits in keystone.
        :type reglimits: dict
        :param projlimits: A dictionary of dictionaries defining per-project
                           limits like {project_id: {resource_name: limit}}.
                           As in reality, only per-project overrides need be
                           provided here; any unmentioned projects or
                           resources will take the registered limit defaults.
        :type projlimits: dict
        """
        self.reglimits = reglimits
        self.projlimits = projlimits

    def get_reglimit_objects(
        self,
        service_id: str | None = None,
        region_id: str | None = None,
        resource_name: str | None = None,
    ) -> list[_registered_limit.RegisteredLimit]:
        registered_limits = []
        for name, value in self.reglimits.items():
            if resource_name and resource_name != name:
                continue

            registered_limit = sdk_fakes.generate_fake_resource(
                _registered_limit.RegisteredLimit,
                resource_name=name,
                default_limit=value,
            )
            registered_limits.append(registered_limit)

        return registered_limits

    def get_projlimit_objects(
        self,
        service_id: str | None = None,
        region_id: str | None = None,
        resource_name: str | None = None,
        project_id: str | None = None,
    ) -> list[_limit.Limit]:
        limits = []
        for proj_id, limit_dict in self.projlimits.items():
            if project_id and project_id != proj_id:
                continue

            for name, value in limit_dict.items():
                if resource_name and resource_name != name:
                    continue

                limit = sdk_fakes.generate_fake_resource(
                    _limit.Limit,
                    resource_name=name,
                    resource_limit=value,
                    project_id=proj_id,
                )
                limits.append(limit)

        return limits

    def setUp(self) -> None:
        super().setUp()

        # We mock our own cached connection to Keystone
        self.mock_conn = mock.MagicMock()
        self.useFixture(
            fixtures.MockPatch(
                'oslo_limit.limit._SDK_CONNECTION', new=self.mock_conn
            )
        )

        # Use a flat enforcement model
        mock_gem = self.useFixture(
            fixtures.MockPatch(
                'oslo_limit.limit.Enforcer._get_enforcement_model'
            )
        ).mock
        mock_gem.return_value = 'flat'

        # Fake keystone endpoint; this can be requested by ID or by name and we
        # need to handle both. First, requests by ID
        fake_endpoint = sdk_fakes.generate_fake_resource(
            _endpoint.Endpoint,
            service_id='service_id',
            region_id='region_id',
        )
        self.mock_conn.get_endpoint.return_value = fake_endpoint

        # Then, requests by name
        self.mock_conn.services.side_effect = _fake_services
        self.mock_conn.get_region.side_effect = _fake_region
        self.mock_conn.endpoints.side_effect = _fake_endpoints

        # Finally, fake the actual limits and registered limits calls
        self.mock_conn.limits.side_effect = self.get_projlimit_objects
        self.mock_conn.registered_limits.side_effect = (
            self.get_reglimit_objects
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_limit._i18n import _


class ProjectOverLimit(Exception):
    def __init__(
        self,
        project_id: str | None,
        over_limit_info_list: list['OverLimitInfo'],
    ) -> None:
        """Exception raised when a project goes over one or more limits

        :param project_id: the project id
        :param over_limit_info_list: list of OverLimitInfo objects
        """
        if not isinstance(over_limit_info_list, list):
            raise ValueError(over_limit_info_list)

        if len(over_limit_info_list) == 0:
            raise ValueError(over_limit_info_list)

        for info in over_limit_info_list:
            if not isinstance(info, OverLimitInfo):
                raise ValueError(over_limit_info_list)

        self.project_id = project_id
        self.over_limit_info_list = over_limit_info_list
        msg = _("Project %(project_id)s is over a limit for %(limits)s") % {
            'project_id': project_id,
            'limits': over_limit_info_list,
        }
        super().__init__(msg)


class OverLimitInfo:
    def __init__(
        self,
        resource_name: str,
        limit: int,
        current_usage: int,
        delta: int,
    ):
        self.resource_name = resource_name
        self.limit = int(limit)
        self.current_usage = int(current_usage)
        self.delta = int(delta)

    def __str__(self) -> str:
        template = (
            "Resource %s is over limit of %s due to "
            "current usage %s and delta %s"
        )
        return template % (
            self.resource_name,
            self.limit,
            self.current_usage,
            self.delta,
        )

    def __repr__(self) -> str:
        return self.__str__()


class SessionInitError(Exception):
    def __init__(self, reason: object) -> None:
        msg = _("Can't initialise OpenStackSDK session: %(reason)s.") % {
            'reason': reason
        }
        super().__init__(msg)
